package org.processmining.streamconformance.regions.plugin.analysis;

import java.net.UnknownHostException;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JPanel;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.connections.ConnectionManager;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.util.Pair;
import org.processmining.models.connections.transitionsystem.TransitionSystemConnection;
import org.processmining.models.graphbased.directed.DirectedGraphElementWeights;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.transitionsystem.AcceptStateSet;
import org.processmining.models.graphbased.directed.transitionsystem.CoverabilityGraph;
import org.processmining.models.graphbased.directed.transitionsystem.StartStateSet;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.streamconformance.regions.gui.panels.ConformanceParameters;
import org.processmining.streamconformance.regions.gui.panels.NetworkConfiguration;
import org.processmining.streamconformance.regions.model.ConformanceTracker;
import org.processmining.streamconformance.regions.plugin.cg.CoverabilityGraphExtenderWithDistance;
import org.processmining.streamconformance.regions.plugin.cg.DummyCoverabilityGraphCreator;
import org.processmining.streamconformance.regions.plugin.cg.ExtendedCoverabilityGraph;

@Plugin(
		name = "Online Conformance - Region-based approach",
		returnLabels = { "A configuration for online conformance checker" },
		returnTypes = { OnlineConformanceConfiguration.class },
		parameterLabels = {
				"Transition System for Conformance",
				"Transition weights",
				"Set of starting states",
				"Set of accepting states",
				"Model", "Marking"
		},
		categories = PluginCategory.Analytics,
		help = "This plugin computes the conformance of a given model with respect to an event streams.",
		userAccessible = true)
public class OnlineConformancePlugin {

	@PluginVariant(requiredParameterLabels = { 0, 1, 2, 3 })
	@UITopiaVariant(
			author = "A. Burattin",
			email = "andbur@dtu.dk",
			affiliation = "DTU")
	public OnlineConformanceConfiguration plugin(UIPluginContext context, CoverabilityGraph originalModel, ExtendedCoverabilityGraph model,
			DirectedGraphElementWeights weights,
			StartStateSet startIds, AcceptStateSet acceptIds) throws UnknownHostException {

		// prepare the configuration panels
		NetworkConfiguration configNetwork = new NetworkConfiguration();
		ConformanceParameters configConformance = new ConformanceParameters();

		List<Pair<String, JPanel>> configurations = new LinkedList<Pair<String, JPanel>>();
		configurations.add(new Pair<String, JPanel>("Network Setup", configNetwork));
		configurations.add(new Pair<String, JPanel>("Conformance Parameters", configConformance));

		// ask the user for the configuration parameters
		InteractionResult result = InteractionResult.NEXT;
		int currentStep = 0;
		int nofSteps = configurations.size();
		boolean configurationOngoing = true;
		while (configurationOngoing && currentStep < nofSteps) {
			Pair<String, JPanel> config = configurations.get(currentStep);
			result = context.showWizard(
					config.getFirst(),
					currentStep == 0,
					currentStep == nofSteps - 1,
					config.getSecond());

			switch (result) {
			case NEXT:
				currentStep++;
				break;
			case PREV:
				currentStep--;
				break;
			case FINISHED:
				configurationOngoing = false;
				break;
			case CANCEL:
				return null;
			default:
				configurationOngoing = false;
				break;
			}
		}
		if (result != InteractionResult.FINISHED) {
			return null;
		}

		// get the final values
		OnlineConformanceConfiguration occ = new OnlineConformanceConfiguration();

		occ.setOriginalModel(originalModel);
		occ.setAddress(configNetwork.getAddress());
		occ.setPort(configNetwork.getPort());
		occ.setModel(model);
		occ.setWeights(weights);
		occ.setStartIds(startIds);
		occ.setAcceptIds(acceptIds);
		occ.setNoMaxParallelInstances(configConformance.getNoMaxParallelInstances());
		occ.setBacktrackingSteps(configConformance.getBacktrackingSteps());
		occ.setNoErorrsPerTrace(configConformance.getNoErrorsPerTrace());
		occ.setNoErrorsMostFrequent(configConformance.getNoMostFrequentErrors());

		return occ;
	}

	@PluginVariant(requiredParameterLabels = { 4, 5 })
	@UITopiaVariant(
			author = "A. Burattin",
			email = "andbur@dtu.dk",
			affiliation = "DTU")
	public OnlineConformanceConfiguration plugin(UIPluginContext context, Petrinet model, Marking state) throws Exception {

		// extend coverability graph
		CoverabilityGraphExtenderWithDistance cge = new CoverabilityGraphExtenderWithDistance();
		Object[] extensionResults = cge.plugin(context, model, state);
		// check if we get something
		if (extensionResults == null) {
			return null;
		}
		// extranct the single results
		CoverabilityGraph originalModel = (CoverabilityGraph) extensionResults[4];
		ExtendedCoverabilityGraph newGraph = (ExtendedCoverabilityGraph) extensionResults[0];
		DirectedGraphElementWeights newWeights = (DirectedGraphElementWeights) extensionResults[1];
		StartStateSet newStartIds = (StartStateSet) extensionResults[2];
		AcceptStateSet newAcceptIds = (AcceptStateSet) extensionResults[3];

		// call the actual plugin
		return plugin(context, originalModel, newGraph, newWeights, newStartIds, newAcceptIds);
	}

	@Plugin(
			name = "Online Conformance Tester",
			parameterLabels = {},
			returnLabels = { "Status" },
			returnTypes = { String.class },
			userAccessible = false)
	@UITopiaVariant(
			author = "A. Burattin",
			email = "andbur@dtu.dk",
			affiliation = "DTU")
	public String test(PluginContext context) throws ConnectionCannotBeObtained {
		Object[] results = (Object[]) DummyCoverabilityGraphCreator.create2(context);
		ExtendedCoverabilityGraph model = (ExtendedCoverabilityGraph) results[0];
		ConnectionManager cm = context.getConnectionManager();
		TransitionSystemConnection tsc = cm.getFirstConnection(TransitionSystemConnection.class, context, model);
		ConformanceTracker ct = new ConformanceTracker(model,
				(DirectedGraphElementWeights) tsc.getObjectWithRole(TransitionSystemConnection.WEIGHTS),
				(StartStateSet) tsc.getObjectWithRole(TransitionSystemConnection.STARTIDS),
				(AcceptStateSet) tsc.getObjectWithRole(TransitionSystemConnection.ACCEPTIDS), 5, 100, 10, 100);

		String[][] toReplay = new String[][] { new String[] { "A", "B", "C", "D" }, new String[] { "A", "D", "C", "B" },
				new String[] { "A", "A", "B", "C", "K" }, new String[] { "B", "B", "C", "D" } };
		int i = 0;
		for (String[] ss : toReplay) {
			System.out.print("Case " + i + ":\t");
			for (String s : ss) {
				System.out.print("`" + s + "' ");
				ct.replay("" + i, s);
			}
			i++;
			System.out.println("");
		}
		for (String c : ct.getHandledCases()) {
			System.out.println("Case " + c + " - final cost: " + ct.get(c).getTotalCost());
		}

		return "Ok";
	}
}
