package org.processmining.streamconformance.regions.plugin.cg;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.connections.ConnectionManager;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.models.connections.transitionsystem.MinimalRegionConnection;
import org.processmining.models.connections.transitionsystem.RegionConnection;
import org.processmining.models.connections.transitionsystem.TransitionSystemConnection;
import org.processmining.models.graphbased.directed.DirectedGraphElementWeights;
import org.processmining.models.graphbased.directed.analysis.ShortestPathFactory;
import org.processmining.models.graphbased.directed.analysis.ShortestPathInfo;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.transitionsystem.AcceptStateSet;
import org.processmining.models.graphbased.directed.transitionsystem.CoverabilityGraph;
import org.processmining.models.graphbased.directed.transitionsystem.StartStateSet;
import org.processmining.models.graphbased.directed.transitionsystem.State;
import org.processmining.models.graphbased.directed.transitionsystem.Transition;
import org.processmining.models.graphbased.directed.transitionsystem.TransitionSystem;
import org.processmining.models.graphbased.directed.transitionsystem.TransitionSystemImpl;
import org.processmining.models.graphbased.directed.transitionsystem.regions.Region;
import org.processmining.models.graphbased.directed.transitionsystem.regions.RegionSet;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.streamconformance.regions.gui.panels.CostsSetup;
import org.processmining.streamconformance.utils.TSUtils;

import cern.colt.matrix.DoubleMatrix1D;
import cern.colt.matrix.impl.SparseDoubleMatrix1D;

public class CoverabilityGraphExtenderWithDistance {

	private UIPluginContext context;
	private TransitionSystemImpl transitionSystem = null;
	private ExtendedCoverabilityGraph coverabilityGraph = null;
	private StartStateSet startIds = new StartStateSet();
	private AcceptStateSet acceptIds = new AcceptStateSet();
	private DirectedGraphElementWeights weights = new DirectedGraphElementWeights();

	private List<String> alphabet = new LinkedList<String>();

	private int costWrongFlow = 1;
	private int costActivityNotInProcess = 1;

	@Plugin(
			name = "Petri net to Extended Coverability graph 2",
			parameterLabels = { "Model", "Marking" },
			returnLabels = {
					"Extended Coverability Graph",
					"Transition weights",
					"Set of starting states",
					"Set of accepting states",
					"Coverability Graph (without extension)"
			},
			returnTypes = {
					ExtendedCoverabilityGraph.class,
					DirectedGraphElementWeights.class,
					StartStateSet.class,
					AcceptStateSet.class,
					CoverabilityGraph.class
			},
			userAccessible = false)
	@UITopiaVariant(
			author = "A. Burattin",
			affiliation = "",
			email = "")
	public Object[] plugin(UIPluginContext context, Petrinet model, Marking state) throws ConnectionCannotBeObtained {

		this.context = context;

		// build coverability graph
		CoverabilityGraph originalGraph = context.tryToFindOrConstructFirstNamedObject(CoverabilityGraph.class,
				"Construct Coverability Graph of a Petri Net", null, null, model, state);
		this.transitionSystem = originalGraph;
		this.alphabet = new LinkedList<>(getProcessAlphabet());

		ConnectionManager cm = context.getConnectionManager();
		TransitionSystemConnection tsc = cm.getFirstConnection(TransitionSystemConnection.class, context, originalGraph);

		StartStateSet originalStartIds = (StartStateSet) tsc.getObjectWithRole(TransitionSystemConnection.STARTIDS);
		AcceptStateSet originalAcceptIds = (AcceptStateSet) tsc.getObjectWithRole(TransitionSystemConnection.ACCEPTIDS);

		// dialog configuration
		CostsSetup cs = new CostsSetup();
		InteractionResult result = context.showWizard("Costs Setup", true, true, cs);

		if (result.equals(InteractionResult.FINISHED)) {
			// perform the actual extension
			Object[] objects = extend(originalGraph, originalStartIds, originalAcceptIds, cs.getCostWrongFlow(),
					cs.getCostActivityNotInProcess());
			// connect the various objects
			context.addConnection(new TransitionSystemConnection(coverabilityGraph, weights, startIds, acceptIds));
			// let the plugin just return the results
			return objects;
		}
		return null;
	}

	public Object[] extend(CoverabilityGraph originalGraph, StartStateSet originalStartIds, AcceptStateSet originalAcceptIds,
			int costWrongFlow,
			int costActivityNotInProcess) {

		this.costWrongFlow = costWrongFlow;
		this.costActivityNotInProcess = costActivityNotInProcess;

		// clone the provided graph in order to create the new
		cloneGraph(originalGraph, originalStartIds, originalAcceptIds);

		// bypass the silent transitions (in this phase we add transitions -> possible violation of determininsm)
		bypassSilent();

		// extend the graph in order to deal with deviations by adding weights
		addWeights();

		// prepare the final stuff to be returned
		Object[] objects = new Object[5];
		objects[0] = coverabilityGraph;
		objects[1] = weights;
		objects[2] = startIds;
		objects[3] = acceptIds;
		objects[4] = originalGraph;
		return objects;
	}

	protected Set<String> getProcessAlphabet() {
		Set<String> alphabet = new HashSet<String>();

		for (Transition t : transitionSystem.getEdges()) {
			if (!TSUtils.isTransitionTau(t)) {
				alphabet.add(t.getLabel());
			}
		}

		return alphabet;
	}

	protected void addNoisyTransition(State source, State target, String activity) {
		coverabilityGraph.addTransition(source.getIdentifier(), target.getIdentifier(), activity);
		Transition transition = coverabilityGraph.findTransition(source.getIdentifier(), target.getIdentifier(), activity);
		ArrayList<Object> list = new ArrayList<Object>(3);
		list.add(transition.getSource().getIdentifier());
		list.add(transition.getTarget().getIdentifier());
		list.add(transition.getIdentifier());
		weights.add(list, costWrongFlow);
	}

	protected void cloneGraph(CoverabilityGraph originalGraph, StartStateSet originalStartIds, AcceptStateSet originalAcceptIds) {
		coverabilityGraph = new ExtendedCoverabilityGraph(originalGraph.getLabel(), costActivityNotInProcess);

		// clone the nodes
		for (State s : originalGraph.getNodes()) {
			coverabilityGraph.addState(s.getIdentifier());
			coverabilityGraph.getNode(s.getIdentifier()).setAccepting(s.isAccepting());
		}
		// clone the transitions
		for (Transition transition : originalGraph.getEdges()) {
			coverabilityGraph.addTransition(transition.getSource().getIdentifier(), transition.getTarget().getIdentifier(),
					transition.getIdentifier());
		}

		// set of starting / accepting nodes
		startIds.addAll(originalStartIds);
		acceptIds.addAll(originalAcceptIds);
	}

	protected void bypassSilent() {
		// bypass silent transitions (in this phase we add transitions -> possible violation of determininsm)
		Set<Transition> processTransitions = new HashSet<Transition>(coverabilityGraph.getEdges());
		for (Transition transition : processTransitions) {
			if (TSUtils.isTransitionTau(transition)) {
				State target = transition.getTarget();
				for (Transition t : coverabilityGraph.getOutEdges(target)) {
					coverabilityGraph.addTransition(transition.getSource().getIdentifier(), t.getTarget().getIdentifier(),
							t.getIdentifier());
				}
			}
		}
	}

	protected void addWeights() {
		for (State state : coverabilityGraph.getNodes()) {
			for (String activity : alphabet) {
				boolean activityFound = false;
				// check all outgoing transitions and verify that from the current state we can execute the activity
				for (Transition transition : coverabilityGraph.getOutEdges(state)) {
					if (transition.getLabel().equals(activity)) {
						activityFound = true;
						break;
					}
				}
				// this activity, from this state represents a deviation, we have to add it, with higher
				if (!activityFound) {
					addNoisyTransition(state, getDeviationTargetState(state, activity), activity);
				}
			}
		}
	}

	protected boolean arcWithSameLabelWithinSameRegion(RegionSet regions, State source, String activity) {
		for (Region region : regions) {
			if (region.contains(source)) {
				// this is one of the regions where the state belongs to
				for (State s : region) {
					for (Transition transition : transitionSystem.getOutEdges(s)) {
						// we have to check for
						if (region.contains(transition.getTarget()) && transition.getLabel().equals(activity)) {
							return true;
						}
					}

				}
			}
		}
		return false;
	}

	protected State getDeviationTargetState(State source, String activity) {
		try {
			RegionSet regions = context.tryToFindOrConstructFirstObject(RegionSet.class, MinimalRegionConnection.class,
					RegionConnection.REGIONS, transitionSystem);

			System.out.print("from state " + source.getLabel() + " --[" + activity + "]--> ");
			if (arcWithSameLabelWithinSameRegion(regions, source, activity)) {
				// if we have an edge labeled with the same as the activity within the same region, then let's add a self loop
				System.out.println(source.getLabel() + " (same region)");
				return source;
			} else {
				StateVector csv = new StateVector(transitionSystem, alphabet, source, activity); // , activity, 0);

				State optimalTarget = null;
				double optimalDistance = Double.MAX_VALUE;

				// check all states with an incoming `activity' edge
				for (Object o : transitionSystem.getStates()) {
					State s = transitionSystem.getNode(o);
					if (stateIsCollectingActivity(s, activity)) {
						StateVector sv = new StateVector(transitionSystem, alphabet, s, activity); // , activity, getDistance(source, s));
						double distance = csv.distance(sv);
						if (distance < optimalDistance) {
							optimalTarget = s;
							optimalDistance = distance;
						}
					}
				}

				if (optimalTarget != null) {
					System.out.println(optimalTarget.getLabel());
					return optimalTarget;
				}
			}

		} catch (ConnectionCannotBeObtained e) {
			e.printStackTrace();
		}

		// by default (should never happen!) let's just return the source state
		System.out.println(source.getLabel() + " (error?)");
		return source;
	}

	private boolean stateIsCollectingActivity(State state, String activity) {
		for (Transition transition : coverabilityGraph.getInEdges(state)) {
			if (transition.getLabel().equals(activity)) {
				return true;
			}
		}
		return false;
	}

	class StateVector {
		private DoubleMatrix1D values = null;
		private TransitionSystem ts = null;
		private State startState = null;
		private List<String> alphabet = null;

		public StateVector(TransitionSystem ts, List<String> alphabet, State state, String targetAction) {
			this.alphabet = alphabet;
			this.values = new SparseDoubleMatrix1D(alphabet.size());
			this.ts = ts;
			
			for (Object s : ts.getStates()) {
				if (ts.getInEdges(ts.getNode(s)).isEmpty()) {
					startState = ts.getNode(s);
					break;
				}
			}

			Set<String> parikh = getPathToStart(state); //getPrecedingLabels(state, new HashSet<State>());
			for (int i = 0; i < alphabet.size(); i++) {
				String s = alphabet.get(i);
				if (s.equals(targetAction)) {
					// System.out.println(s + " = t; ");
					values.set(i, 0);
				} else {
					if (parikh.contains(s)) {
						// System.out.println(s + " = " + 1 + "; ");
						values.set(i, 1);
					} else {
						// System.out.println(s + " = " + 0 + "; ");
						values.set(i, 0);
					}
				}
			}
			// System.out.println("");
			// Set<String> incoming = new HashSet<String>();
			// Set<String> outgoing = new HashSet<String>();
			//
			// for (Transition t : transitionsystem.getInEdges(state)) {
			// incoming.add(t.getLabel());
			// }
			//
			// for (Transition t : transitionsystem.getOutEdges(state)) {
			// outgoing.add(t.getLabel());
			// }
			//
			// for (int i = 0; i < alphabet.size() - 1; i++) {
			// String s = alphabet.get(i);
			// if (s.equals(targetAction)) {
			// values.set(i, targetDistance);
			// } else {
			// if (incoming.contains(s)) {
			// values.set(i, -1);
			// } else if (outgoing.contains(s)) {
			// values.set(i, +1);
			// } else {
			// values.set(i, 0);
			// }
			// }
			// }
		}
		
		private Set<String> getPathToStart(State fromState) {
			ShortestPathInfo<State, Transition> info = ShortestPathFactory.calculateAllShortestDistanceDijkstra(ts);
			Set<String> result = new HashSet<String>();
			State prev = null;
			for (State s : info.getShortestPath(startState, fromState)) {
				if (prev != null) {
					result.add(TSUtils.getConnection(prev, s).getLabel());
				}
				prev = s;
			}
			return result;
		}

//		private Set<String> getPrecedingLabels(State s, Set<State> visited) {
//			Set<String> result = new HashSet<String>();
//			for (Transition t : ts.getInEdges(s)) {
//				State source = t.getSource();
//				if (!visited.contains(source)) {
//					visited.add(s);
//					result.add(t.getLabel());
//					result.addAll(getPrecedingLabels(source, visited));
//				}
//			}
//			return result;
//		}

		@Override
		public String toString() {
			String s = "[";
			for (int i = 0; i < values.size(); i++) {
				s += alphabet.get(i) + " = " + values.get(i) + ", ";
			}
			s += "]";
			return s;
		}

		public double distance(StateVector v2) {
			DoubleMatrix1D a = this.values;
			DoubleMatrix1D b = v2.values;
			double distance = 1d - (a.zDotProduct(b) / Math.sqrt(a.zDotProduct(a) * b.zDotProduct(b)));
			// System.out.println(v1 + " - " + v2 + " = " + distance);
			return distance;
		}
	}
}
