package org.processmining.streamconformance.utils;

import java.io.BufferedReader;
import java.io.File;
import java.io.InputStreamReader;

import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.util.Pair;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.impl.PetrinetFactory;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.pnml.exporting.PnmlExportNetToPNML;
import org.processmining.plugins.pnml.importing.PnmlImportNet;

public class PetrinetHelper {

	public static Pair<Petrinet, Marking> unfold(PluginContext context, Petrinet net) throws Exception {
		File pnmlOriginal = File.createTempFile("petrinet", "original");
		String pnmlUnfolded = pnmlOriginal.getAbsolutePath() + "-unfolded";
		
		PnmlExportNetToPNML exporter = new PnmlExportNetToPNML();
		exporter.exportPetriNetToPNMLFile(context, net, pnmlOriginal);
		
		ProcessBuilder pb = new ProcessBuilder("java", "-jar", "./lib/unfolder-v4.jar", pnmlOriginal.getAbsolutePath(), pnmlUnfolded);
		Process proc = pb.start();
		BufferedReader reader = new BufferedReader(new InputStreamReader(proc.getInputStream()));
		while (reader.readLine() != null) { }
		proc.waitFor();
		
		
		Petrinet netUnfolded = context.tryToFindOrConstructFirstNamedObject(
				Petrinet.class,
				PnmlImportNet.class.getAnnotation(Plugin.class).name(),
				null,
				null,
				pnmlUnfolded);
		InitialMarkingConnection imc = context.getConnectionManager().getFirstConnection(InitialMarkingConnection.class, context, netUnfolded);
		
		pnmlOriginal.deleteOnExit();
		return new Pair<Petrinet, Marking>(
				(Petrinet) imc.getObjectWithRole("Net"),
				(Marking) imc.getObjectWithRole("Marking"));
	}
	
	public static Pair<Petrinet, Marking> computeDual(PluginContext context, Petrinet net) {
		Petrinet dualNet = PetrinetFactory.clonePetrinet(net);
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : dualNet.getEdges()) {
			dualNet.removeEdge(e);
			if (e.getTarget() instanceof Place && e.getSource() instanceof org.processmining.models.graphbased.directed.petrinet.elements.Transition) {
				dualNet.addArc((Place) e.getTarget(), (org.processmining.models.graphbased.directed.petrinet.elements.Transition) e.getSource());
			} else {
				dualNet.addArc((org.processmining.models.graphbased.directed.petrinet.elements.Transition) e.getTarget(), (Place) e.getSource());
			}
		}
		
		Marking initial = new Marking();
		for(Place p : dualNet.getPlaces()) {
			if (dualNet.getInEdges(p).isEmpty()) {
				initial.add(p);
			}
		}

		context.getProvidedObjectManager().createProvidedObject("dualnet", dualNet, context);
		context.getProvidedObjectManager().createProvidedObject("dualnet-markin", initial, context);
		context.getConnectionManager().addConnection(new InitialMarkingConnection(dualNet, initial));
		
		return new Pair<Petrinet, Marking>(dualNet, initial);
	}
	
	public static boolean isWorkflowNet(Petrinet net) {
		boolean hasSourcePlace = false;
		boolean hasSinkPlace = false;
		for (Place p : net.getPlaces()) {
			hasSourcePlace = hasSourcePlace || net.getInEdges(p).isEmpty();
			hasSinkPlace = hasSinkPlace || net.getOutEdges(p).isEmpty();
		}
		return (hasSourcePlace && hasSinkPlace);
	}
}
